(*
 * LGPL-3.0 Linking Exception
 *
 * Copyright (c) 2020 Joan Thibault (joan.thibault@irisa.fr)
 *
 * GuaCaml : Generic Unspecific Algorithmic in OCaml
 *
 * RLE : Run-Length Encoding
 *)

type 'a t = (int * 'a) list

let rec reduced_rec (x:'a) : 'a t -> bool = function
  | [] -> true
  | (ny, y)::t -> ny >= 1 && x <> y && reduced_rec y t

let reduced : 'a t -> bool = function
  | [] -> true
  | (nx, x)::t -> nx >= 1 && reduced_rec x t

let cons ?(n=1) (x:'a) (t:'a t) : 'a t =
  assert(n >= 0);
  if n = 0
  then t
  else match t with
    | [] -> [(n, x)]
    | (ny, y)::t' ->
      if x = y
      then ((n+ny, x)::t')
      else ((n   , x)::t )

let hdtl : 'a t -> 'a * 'a t = function
  | [] -> (failwith "[GuaCaml.RLE] error")
  | (nx, x)::t ->
    if nx = 1 then (x, t) else (x, (pred nx, x)::t)

let rec hdtl_nth ?(carry=[]) (n:int) (t:'a t) : 'a t * 'a t =
  if n = 0
  then (List.rev carry, t)
  else (
    assert(n > 0)
    match t with
    | [] -> (failwith "[GuaCaml.RLE] error")
    | (nx, x)::t -> hdtl_nth
  else function
  | [] -> assert(n = 0); (List.rev carry, [])
  | (nx, x)::t -> if

